/**
 *  on(HTMLElement, eventType, event) 监听原生事件
 *  off(HTMLElement, eventType, event) 移除原生事件
 *  once(HTMLElement, eventType, event) 监听原生事件，但只触发一次
 *  getStyle(HTMLElement, styleName) 获取目标样式属性
 *  setStyle(HTMLElement, styleName, value) 设置标样式属性
 */

const isServer = typeof window == undefined;
const SPECIAL_CHARS_REGEXP = /([\:\-\_]+(.))/g;
const MOZ_HACK_REGEXP = /^moz([A-Z])/;
const ieVersion = isServer ? 0 : Number(document.documentMode);

const camelCase = function (name) {
  return name
    .replace(SPECIAL_CHARS_REGEXP, function (_, separator, letter, offset) {
      return offset ? letter.toUpperCase() : letter;
    })
    .replace(MOZ_HACK_REGEXP, 'Moz$1');
};

export const isClosest = function (childElement, parentElement) {
  if (
    parentElement.contains &&
    parentElement.contains(childElement) &&
    parentElement != childElement
  ) {
    return true;
  } else if (
    parentElement.compareDocumentPosition &&
    parentElement.compareDocumentPosition(childElement) & 16 &&
    parentElement != childElement
  ) {
    return true;
  }
  return false;
};

export const on = (function () {
  if (!isServer && document.addEventListener) {
    return function (element, event, handler) {
      if (element && event && handler) {
        element.addEventListener(event, handler, false);
      }
    };
  } else {
    return function (element, event, handler) {
      if (element && event && handler) {
        element.attachEvent('on' + event, handler);
      }
    };
  }
})();

export const off = (function () {
  if (!isServer && document.removeEventListener) {
    return function (element, event, handler) {
      if (element && event) {
        element.removeEventListener(event, handler, false);
      }
    };
  } else {
    return function (element, event, handler) {
      if (element && event) {
        element.detachEvent('on' + event, handler);
      }
    };
  }
})();

export const once = function (el, event, fn) {
  const listener = function () {
    if (fn) {
      fn.apply(this, arguments);
    }
    off(el, event, listener);
  };
  on(el, event, listener);
};

export const getStyle =
  ieVersion < 9
    ? function (element, styleName) {
        if (isServer) return;
        if (!element || !styleName) return null;
        styleName = camelCase(styleName);
        if (styleName === 'float') {
          styleName = 'styleFloat';
        }
        try {
          switch (styleName) {
            case 'opacity':
              try {
                return element.filters.item('alpha').opacity / 100;
              } catch (e) {
                return 1.0;
              }
            default:
              return element.style[styleName] || element.currentStyle
                ? element.currentStyle[styleName]
                : null;
          }
        } catch (e) {
          return element.style[styleName];
        }
      }
    : function (element, styleName) {
        if (isServer) return;
        if (!element || !styleName) return null;
        styleName = camelCase(styleName);
        if (styleName === 'float') {
          styleName = 'cssFloat';
        }
        try {
          var computed = document.defaultView.getComputedStyle(element, '');
          return element.style[styleName] || computed
            ? computed[styleName]
            : null;
        } catch (e) {
          return element.style[styleName];
        }
      };

export function setStyle(element, styleName, value) {
  if (!element || !styleName) return;

  if (typeof styleName === 'object') {
    for (var prop in styleName) {
      if (styleName.hasOwnProperty(prop)) {
        setStyle(element, prop, styleName[prop]);
      }
    }
  } else {
    styleName = camelCase(styleName);
    if (styleName === 'opacity' && ieVersion < 9) {
      element.style.filter = isNaN(value)
        ? ''
        : 'alpha(opacity=' + value * 100 + ')';
    } else {
      element.style[styleName] = value;
    }
  }
}

// ===================
export const throttle = function (delay, noTrailing, callback, debounceMode) {
  let timeoutID;
  let lastExec = 0;
  if (typeof noTrailing !== 'boolean') {
    debounceMode = callback;
    callback = noTrailing;
    noTrailing = undefined;
  }

  function wrapper() {
    let self = this;
    let elapsed = Number(new Date()) - lastExec;
    let args = arguments;

    function exec() {
      lastExec = Number(new Date());
      callback.apply(self, args);
    }

    function clear() {
      timeoutID = undefined;
    }

    if (debounceMode && !timeoutID) {
      exec();
    }

    if (timeoutID) {
      clearTimeout(timeoutID);
    }

    if (debounceMode === undefined && elapsed > delay) {
      exec();
    } else if (noTrailing !== true) {
      timeoutID = setTimeout(
        debounceMode ? clear : exec,
        debounceMode === undefined ? delay - elapsed : delay
      );
    }
  }
  return wrapper;
};

export const debounce = function (delay, atBegin, callback) {
  return callback === undefined
    ? throttle(delay, atBegin, false)
    : throttle(delay, callback, atBegin !== false);
};

const generateId = function () {
  return Math.floor(Math.random() * 10000);
};

export default {
  throttle,
  debounce,
  generateId,
  on,
  off,
  once,
  getStyle,
  setStyle,

  extend: function () {
    var options,
      name,
      src,
      copy,
      copyIsArray,
      clone,
      target = arguments[0] || {},
      i = 1,
      length = arguments.length,
      deep = false;
    if (typeof target === 'boolean') {
      deep = target;
      target = arguments[1] || {};
      i = 2;
    }
    if (typeof target !== 'object' && !this.isFunction(target)) {
      target = {};
    }
    if (length === i) {
      target = this;
      --i;
    }
    for (; i < length; i++) {
      if ((options = arguments[i]) != null) {
        for (name in options) {
          src = target[name];
          copy = options[name];
          if (src === copy) {
            continue;
          }
          if (
            deep &&
            copy &&
            (this.isPlainObject(copy) || (copyIsArray = this.isArray(copy)))
          ) {
            if (copyIsArray) {
              copyIsArray = false;
              clone = src && this.isArray(src) ? src : [];
            } else {
              clone = src && this.isPlainObject(src) ? src : {};
            }
            target[name] = this.extend(deep, clone, copy);
          } else if (copy !== undefined) {
            target[name] = copy;
          }
        }
      }
    }
    return target;
  },
  freeze(obj) {
    const that = this;
    Object.freeze(obj);
    Object.keys(obj).forEach((key, value) => {
      if (that.isObject(obj[key])) {
        this.freeze(obj[key]);
      }
    });
    return obj;
  },

  getKeyValue(obj, keypath) {
    if (!this.isObject(obj)) {
      return null;
    }
    let array = null;
    if (this.isArray(keypath)) {
      array = keypath;
    } else if (this.isString(keypath)) {
      array = keypath.split('.');
    }
    if (array == null || array.length == 0) {
      return null;
    }
    let value = null;
    let key = array.shift();
    const keyTest = key.match(new RegExp('^(\\w+)\\[(\\d+)\\]$'));
    if (keyTest) {
      key = keyTest[1];
      let index = keyTest[2];
      value = obj[key];
      if (this.isArray(value) && value.length > index) {
        value = value[index];
      }
    } else {
      value = obj[key];
    }

    if (array.length > 0) {
      return this.getKeyValue(value, array);
    }
    return value;
  },
  setKeyValue(obj, keypath, value, orignal) {
    if (!this.isObject(obj)) {
      return false;
    }
    let array = null;
    if (this.isArray(keypath)) {
      array = keypath;
    } else if (this.isString(keypath)) {
      array = keypath.split('.');
      orignal = obj;
    }
    if (array == null || array.length == 0) {
      return false;
    }
    let children = null;
    let index = 0;
    let key = array.shift();
    const keyTest = key.match(new RegExp('^(\\w+)\\[(\\d+)\\]$'));
    if (keyTest) {
      key = keyTest[1];
      index = keyTest[2];
      children = obj[key];
      if (this.isArray(children) && children.length > index) {
        if (array.length > 0) {
          return this.setKeyValue(children[index], array, value, orignal);
        }
        children[index] = value;
      }
    } else {
      if (array.length > 0) {
        return this.setKeyValue(obj[key], array, value, orignal);
      }
      obj[key] = value;
    }
    return orignal;
  },
  padLeft(str, size) {
    var s = '00000' + str;
    return s.substr(s.length - size);
  },

  arrayEquals(a, b) {
    if (a === b) return true;
    if (!(a instanceof Array)) return false;
    if (!(b instanceof Array)) return false;
    if (a.length !== b.length) return false;
    for (let i = 0; i !== a.length; ++i) {
      if (a[i] !== b[i]) return false;
    }
    return true;
  },

  trimAll(string) {
    return (string || '').replace(/\s+/g, '');
  },
  cut(string, len) {
    if (string.length > len) {
      return string.substring(0, len) + '...';
    } else {
      return string;
    }
  },
  escape(string) {
    return string
      ? string
          .replace(/&/g, '&amp;')
          .replace(/</g, '&lt;')
          .replace(/>/g, '&gt;')
          .replace(/"/g, '&quot;')
          .replace(/'/g, '&#039;')
      : '';
  },
  beginWith(string, tag) {
    return string.substring(0, tag.length) == tag;
  },
  endWidth() {
    return string.substring(string.length - tag.length) == tag;
  }
};
