import { isArray, isPlainObject, toRawType } from './type';

const EnableCopyList =
  'Number,String,Boolean,Null,Undefined,Symbol,BigInt,Object,Array'.split(',');

/**
 * 复制一份数据，仅支持基础类型和[object Object]类型以及[object Array]类型
 * 复杂的类型会给出警告，通过将第二个参数设置为 true 则关闭警告
 * @param source
 * @param silence 不打印警告
 * @example
 *  copy(1) => 1
 *  copy('2') => 2
 *  copy(null) => null
 *  copy(undefined) => undefined
 *  copy([1,3]) => [1,3]
 *  copy(Symbol(1)) => Symbol(1)
 *  copy(() => {}) => 产生警告
 *  copy(new Map()) => 产生警告
 *  ... 不一一列举
 */
export default function copy<T = unknown>(source: T, silence = false): T {
  let copyOne;

  if (isPlainObject(source)) {
    copyOne = {} as Record<any, any>;
    for (const prop in source) {
      copyOne[prop] = copy(source[prop], silence);
    }
  } else if (isArray(source)) {
    copyOne = [] as any[];
    for (const item of <any[]>source) {
      copyOne.push(copy(item, silence));
    }
  } else {
    const rawType = toRawType(source);
    if (!EnableCopyList.includes(rawType) && !silence) {
      console.warn(
        `The [${rawType}] type property of the current target '${source}' cannot be copied.`
      );
    }
    copyOne = source;
  }
  return copyOne;
}
